// Copyright 2021 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

syntax = "proto3";

package chip.rpc;

message Empty {}

message Button {
  // 0=FUNCTION 1=LIGHTING 2=THREAD_START 3=BLE_START
  uint32 idx = 1;

  // 0=release 1=push
  uint32 action = 2;
}

service LightingService {
  // Pigweed RPC button event
  // See more details from
  // https://github.com/project-chip/connectedhomeip/tree/master/examples/lighting-app/nrfconnect
  rpc ButtonEvent(Button) returns (Empty) {}
}
